static  const char rcsid[] = "$Id: conf.c,v 1.5 2004/05/01 15:28:20 dengxf Exp $";

/*
 * Copyright (c) 2003-2004 Deng XueFeng <dsnofe@hotmail.com>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *    This product includes software developed by Deng XueFeng.
 * 4. Neither the name of the author nor the names of any co-contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY DENG XUEFENG AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>

#include "log.h"
#include "checkip.h"

extern int pop3d_port;
extern int max_errs;
extern int multi_process;
extern int max_backlog;
extern int max_client;
extern int max_timeout;
extern char serv_ip[16];

extern int errno;

/*
 *  -1 -- ϵͳ
 *   0 -- Ӵ
 *   1 -- ȷ
 */

int add_conf(register char *const conf_line)
{
    int i;
    register char *start, *end;

    start = conf_line;
    while ((*start) && (*start == ' ')) {
        start++;
    }
    if (*start == '#') {
        return (0);
    }
    if ((start = strstr(conf_line, "Client_Max_Errors")) != NULL) {
        start += sizeof("Client_Max_Errors");
        while (*start == ' ') {
            start++;
        }
        end = start;
        while ((*end != ' ') && (*end != '\0')){
            end++;
        }
        *end = '\0';
        i = atoi(start);
        if (i > 0) {
            max_errs = i;
        }
        return (1);
    }
    if ((start = strstr(conf_line, "Client_Time_Out")) != NULL) {
        start += sizeof("Client_Time_Out");
        while (*start == ' ') {
            start++;
        }
        end = start;
        while ((*end != ' ') && (*end != '\0')){
            end++;
        }
        *end = '\0';
        i = atoi(start);
        if (i > 0) {
            max_timeout = i * 1000;
        }
        return (1);
    }
    if ((start = strstr(conf_line, "Per_Process_Max_Client")) != NULL) {
        start += sizeof("Per_Process_Max_Client");
        while (*start == ' ') {
            start++;
        }
        end = start;
        while ((*end != ' ') && (*end != '\0')){
            end++;
        }
        *end = '\0';
        i = atoi(start);
        if (i > 0) {
            max_client = i;
        }
        return (1);
    }
    if ((start = strstr(conf_line, "Multi_Process")) != NULL) {
        start += sizeof("Multi_Process");
        while (*start == ' ') {
            start++;
        }
        end = start;
        while ((*end != ' ') && (*end != '\0')){
            end++;
        }
        *end = '\0';
        i = atoi(start);
        if (i > 0) {
            multi_process = i;
        }
        return (1);
    }
    if ((start = strstr(conf_line, "Max_Backlog")) != NULL) {
        start += sizeof("Max_Backlog");
        while (*start == ' ') {
            start++;
        }
        end = start;
        while ((*end != ' ') && (*end != '\0')){
            end++;
        }
        *end = '\0';
        i = atoi(start);
        if (i > 0) {
            max_backlog = i;
        }
        return (1);
    }
    if ((start = strstr(conf_line, "Server_Port")) != NULL) {
        start += sizeof("Server_Port");
        while (*start == ' ') {
            start++;
        }
        end = start;
        while ((*end != ' ') && (*end != '\0')){
            end++;
        }
        *end = '\0';
        i = atoi(start);
        if (i > 0) {
            pop3d_port = i;
        }
        return (1);
    }
    if ((start = strstr(conf_line, "Server_Address")) != NULL) {
        start += sizeof("Server_Address");
        while (*start == ' ') {
            start++;
        }
        end = start;
        while ((*end != ' ') && (*end != '\0')){
            end++;
        }
        *end = '\0';
        if (checkip(start, serv_ip) != 1) {
            INFO("Warning: Invalid IP Address: [%s]\n", start);
            memset(serv_ip, 0x00, sizeof(serv_ip));
            return (0);
        }
        return (1);
    }
    return (0);
}

/*
 * ʼļ
 * ֵ:
 *  0 -- 
 *  1 -- ȷ
 */

int parse_conf(const char *cfg_file)
{
    struct stat stat_buf;
    register char *save;
    register char *start;
    register char *end;
    register short int fd_cfg_file;

    fd_cfg_file = open(cfg_file, O_RDONLY);
    if (fd_cfg_file == -1) {
        INFO("ERROR: open(2) %s failed: %s\n", cfg_file, strerror(errno));
        return (0);
    }
    if ((stat(cfg_file, &stat_buf)) == -1) {
        INFO("ERROR: stat(2) %s failed: %s\n", cfg_file, strerror(errno));
        close(fd_cfg_file);
        return (0);
    }
    save = mmap(0, stat_buf.st_size, PROT_READ|PROT_WRITE, MAP_PRIVATE, fd_cfg_file, 0L);
    close(fd_cfg_file);
    if (save == MAP_FAILED) {
        INFO("ERROR: mmap(2) %s failed: %s\n", cfg_file, strerror(errno));
        return (0);
    }
    start = save;
    while ((*start == '\r') || (*start == '\n')) {
        start++;
    }
    end = save;
    while (end < stat_buf.st_size + save) {
        if ((*end == '\r') || (*end == '\n')) {
            *end++ = '\0';
            if (add_conf(start) == -1) {
                return (0);
            }
            while (end < stat_buf.st_size + save) {
                if ((*end == '\r') || (*end == '\n')) {
                    end++;
                } else {
                    break;
                }
            }
            start = end;
        } else {
            end++;
        }
    }
    munmap(save, stat_buf.st_size);
    return (1);
}
