static  const char rcsid[] = "$Id: main.c,v 1.7 2004/05/06 14:02:24 dengxf Exp $";

/*
 * Copyright (c) 2003-2004 Deng XueFeng <dsnofe@hotmail.com>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *    This product includes software developed by Deng XueFeng.
 * 4. Neither the name of the author nor the names of any co-contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY DENG XUEFENG AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <syslog.h>
#include <stdarg.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/resource.h>

#include <sys/types.h>
#include <sys/wait.h>


#include "log.h"
#include "sig.h"
#include "listen.h"
#include "pop3d.h"
#include "mydaemon.h"
#include "conf.h"

char config_file[] = "/usr/local/etc/missey/pop3d.conf";

int pop3d_port = 6110;
int max_errs = 5;
int multi_process = 5;
int max_backlog = 1024;
int max_client = 1000;
int max_timeout = 10 * 1000;
char serv_ip[16];

int debug_flag = 0;
int pop3d_fd = 0;
int cur_process = 0;

extern int errno;

static int pop3d_child()
{
    int pid;

    pid = fork();
    if (pid < 0) {
        INFO("missey::pop3d::main: fork(2) Failed first time: %s", strerror(errno));
        return (0);
    } else if (pid == 0) {
        pop3d();
        return (1);
    } else {
        //INFO("missey::pop3d::process[%d] started", pid);
        return (1);
    }
}

static void usage(void)
{
    printf("Usage: missey [-c configfile] [-d]\n");
    printf("       -c   Specified configfile, default is /usr/local/etc/missey/pop3d.conf\n");
    printf("       -d   Debug Mod\n");
    exit(EXIT_FAILURE);
}

int main(int argc, char **argv)
{
    char opt, cfg_file[256];
    struct sigaction act;
    int err;

    (void) memset(serv_ip, 0, sizeof(serv_ip));
    (void) memset(cfg_file, 0, sizeof(config_file));
    while ((opt = getopt(argc, argv, "c:d")) != -1) {
        switch(opt) {
            case 'c':
                strlcpy(cfg_file, optarg, sizeof(cfg_file));
                cfg_file[255] = '\0';
                break;
            case 'd':
                debug_flag = 1;
                break;
            default :
                usage();
        }
    }
    if (cfg_file[0] == '\0') {
        strlcpy(cfg_file, config_file, sizeof(cfg_file));
    }
    if (parse_conf(cfg_file) != 1) {
        usage();
    }
    if (debug_flag == 1) {
        printf("DEBUG: MOD\n");
        printf("DEBUG: pop3d_port: [%d]\n", pop3d_port);
        printf("DEBUG: max_errs: [%d]\n", max_errs);
        printf("DEBUG: multi_process: [%d]\n", multi_process);
        printf("DEBUG: max_backlog: [%d]\n", max_backlog);
        printf("DEBUG: max_client: [%d]\n", max_client);
        printf("DEBUG: max_timeout: [%d]\n", max_timeout);
        printf("DEBUG: serv_ip: [%s]\n", serv_ip);
    }
    pop3d_fd = newlisten(&err, serv_ip, pop3d_port, 10, max_backlog);
    switch (pop3d_fd) {
        case ERROR_PORT:
            printf("-ERR INVALID PORT: [%d].\n", pop3d_port);
            exit(EXIT_FAILURE);
        case ERROR_IP:
            printf("-ERR INVALID IP ADDRESS: [%s].\n", serv_ip);
            exit(EXIT_FAILURE);
        case ERROR_SOCKET:
            printf("-ERR CREATE SOCKET FAILED: [%s].\n", strerror(err));
            exit(EXIT_FAILURE);
        case ERROR_REUSE:
            printf("-ERR SETSOCKOPT REUSEADDR FAILED: [%s].\n", strerror(err));
            exit(EXIT_FAILURE);
        case ERROR_LINGER:
            printf("-ERR SETSOCKOPT LINGER FAILED: [%s].\n", strerror(err));
            exit(EXIT_FAILURE);
        case ERROR_BIND:
            printf("-ERR BIND SOCKET FAILED: [%s].\n", strerror(err));
            exit(EXIT_FAILURE);
        case ERROR_LISTEN:
            printf("-ERR LISTEN SOCKET FAILED: [%s].\n", strerror(err));
            exit(EXIT_FAILURE);
        default:
            break;
    }
    if (debug_flag == 0) {
        mydaemon();
    }
    sigemptyset(&act.sa_mask);
    act.sa_flags = 0;
    act.sa_handler = sig_exit_handler;
    sigaction(SIGTERM, &act, NULL);
    sigaction(SIGQUIT, &act, NULL);
    sigaction(SIGILL, &act, NULL);
    sigaction(SIGABRT, &act, NULL);
    sigaction(SIGFPE, &act, NULL);
    sigaction(SIGBUS, &act, NULL);
    sigaction(SIGSEGV, &act, NULL);
    act.sa_handler = sig_handler;
    sigaction(SIGPIPE,&act,NULL);
    cur_process = 0;

    if (debug_flag == 1) {
        pop3d();
        return (0);
    }
    while (cur_process < multi_process) {
        if (pop3d_child() == 0) {
            kill(0, SIGINT); // KILL GROUP Process
        }
        cur_process++;
    }
    act.sa_handler = sig_chld_handler;
    sigaction(SIGCHLD,&act,NULL);
    while(1) {
        sleep(30);
        if (cur_process < multi_process) {
            if (pop3d_child() == 1) {
                cur_process++;
            }
        }
    }
    INFO("EXIT SUCCESSFUL\n");
    return (0);
}
