static  const char rcsid[] = "$Id: pop3d.c,v 1.8 2004/05/06 14:02:24 dengxf Exp $";

/*
 * Copyright (c) 2003-2004 Deng XueFeng <dsnofe@hotmail.com>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *    This product includes software developed by Deng XueFeng.
 * 4. Neither the name of the author nor the names of any co-contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY DENG XUEFENG AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#include <dirent.h>
#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <netinet/in.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <sys/event.h>
#include <sys/socket.h>

#include "pop3d.h"
#include "const.h"
#include "auth.h"
#include "plcat.h"
#include "log.h"

static int kq = 0;
static int online = 0;
static int used_afds = 0;
static int all_fid = 64;
static int used_fid = 0;
static int all_cudb = 128;
static int used_audb = 0;
static int used_cudb = 0;
static int *fid = NULL;
static struct kevent *afds = NULL;
static struct kevent *aevs = NULL;
static UDB *audb = NULL;
static UDB **cudb = NULL;

extern int max_errs;
extern int max_client;
extern int max_timeout;
extern int pop3d_fd;

static int kq_init(void)
{
    kq = kqueue();
    if (kq == -1) {
        INFO("kq_init: kqueue(2) failed: %s\n", strerror(errno));
        return (-1);
    }
    fid = (int *) malloc(all_fid * sizeof(int));
    afds = (struct kevent *) malloc(max_client * sizeof(struct kevent));
    aevs = (struct kevent *) malloc(max_client * sizeof(struct kevent));
    audb = (UDB *) malloc(max_client * sizeof(UDB));
    cudb = (UDB **) malloc(all_cudb * sizeof(UDB *));
    if ((fid == (int *) 0) ||
            (afds == (struct kevent *) 0) ||
            (aevs == (struct kevent *) 0) ||
            (audb == (UDB *) 0) ||
            (cudb == (UDB **) 0)) {
        INFO("kq_init: malloc(2) failed at kq_init: %s\n", strerror(errno));
        return (-1);
    }
    (void) memset(fid, 0, all_fid * sizeof(int));
    (void) memset(afds, 0, max_client * sizeof(struct kevent));
    (void) memset(aevs, 0, max_client * sizeof(struct kevent));
    (void) memset(audb, 0, max_client * sizeof(UDB));
    (void) memset(cudb, 0, all_cudb * sizeof(UDB *));
    return (0);
}

static void exitfree(void)
{
    if (fid != NULL) {
        free(fid);
        fid = NULL;
    }
    if (afds != NULL) {
        free(afds);
        afds = NULL;
    }
    if (aevs != NULL) {
        free(aevs);
        aevs = NULL;
    }
    if (audb != NULL) {
        free(audb);
        audb = NULL;
    }
    if (cudb != NULL) {
        free(cudb);
        cudb = NULL;
    }
}

static void update_fid(const int id)
{
    int *p;

    if (used_fid == all_fid) {
        all_fid <<= 1;
        p = (int *) realloc(fid, all_fid * sizeof(int));
        if (p == NULL) {
            INFO("update_fid: realloc(2) failed at update_fid: %s\n", strerror(errno));
            exitfree();
            _exit(EXIT_FAILURE);
        }
        (int *) fid = (int *) p;
    }
    *(fid + used_fid++) = id;
}

static void update_udb(UDB *const addr)
{
    UDB **p;

    if (used_cudb == all_cudb) {
        all_cudb <<= 1;
        p = (UDB **) realloc(cudb, all_cudb * sizeof(UDB *));
        if (p == NULL) {
            INFO("update_udb: realloc(2) failed at update_fid: %s\n", strerror(errno));
            exitfree();
            _exit(EXIT_FAILURE);
        }
        (UDB **) cudb = (UDB **) p;
    }
    *(cudb + used_cudb++) = addr;
}

static ssize_t Write(const int fd, const void *const buf, const size_t n)
{
    ssize_t nwritten;

    nwritten = write(fd, buf, n);
    if (nwritten != n) {
        INFO("Write: write(2) failed at Write: %s\n", strerror(errno));
        return (-1);
    }
    return (nwritten);
}

static ssize_t Awrite(const int fd, const void *const buf, const size_t n)
{
    ssize_t nleft;
    ssize_t nwritten;
    const char *ptr;

    nleft = n;
    ptr = buf;
    while (nleft > 0) {
        if ((nwritten = write(fd, ptr, nleft)) <= 0) {
            if ((errno == EINTR) || (errno == EAGAIN)) {
                nwritten = 0;
            } else {
                INFO("Awrite: write(2) failed at Awrite: %s\n", strerror(errno));
                return (-1);
            }
        }
        nleft -= nwritten;
        ptr += nwritten;
    }
    return (n);
}

static void cli_accept(const struct kevent *const kdp)
{
    int cli_fd;
    UDB *udbp;
    socklen_t tmpaddrlen;
    struct sockaddr_in address;

    tmpaddrlen = sizeof(struct sockaddr_in);
    cli_fd = accept(pop3d_fd, (struct sockaddr *) &address, &tmpaddrlen);
    if (cli_fd < 0) {
        INFO("cli_accept: accept(2) failed at cli_accept: %s\n", strerror(errno));
        return;
    } else {
        if (online == max_client) {
            Write(cli_fd, msg_max_online, sizeof(msg_max_online) - 1);
            close(cli_fd);
            return;
        }
        udbp = (used_cudb == 0) ? &audb[used_audb++] : cudb[--used_cudb];
        udbp->free = (udbp->free == DEF_UDB_FREEED) ? DEF_UDB_FREEED_THEN_USED : DEF_UDB_USED;
        udbp->welmsg = 0;
        udbp->login = 0;
        udbp->c_status = DEF_CACHE_UNDONE;
        udbp->cmd = 0;
        udbp->cmderr = 0;
        udbp->curerrs = 0;
        udbp->maxid = 0;
        udbp->upid = 0;
        udbp->ucsize = 0;
        udbp->a_id = 0;
        udbp->offset = 0;
        udbp->mail = NULL;
        udbp->c_all = NULL;
        udbp->c_updated = NULL;
        udbp->username[0] = '\0';
        udbp->maildir[0] = '\0';
        EV_SET(&afds[used_afds++], cli_fd, EVFILT_WRITE, EV_ADD | EV_ONESHOT, 0, 0, udbp);
        online++;
    }
}

static void myfree(const struct kevent *const kdp)
{
    UDB *t_udb;

    t_udb = (UDB *) kdp->udata;
    if (t_udb->c_all != NULL) {
        free(t_udb->c_all);
    }
    if (t_udb->c_updated != NULL) {
        free(t_udb->c_updated);
    }
    EV_SET(&afds[used_afds++], kdp->ident, EVFILT_TIMER, EV_DELETE, 0, 0, NULL);
    update_udb(kdp->udata);
    update_fid(kdp->ident);
    t_udb->free = DEF_UDB_FREEED;
    close(kdp->ident);
    online--;
}

static void check_errs(const struct kevent *const kdp)
{
    UDB *t_udb;

    t_udb = (UDB *) kdp->udata;
    if (t_udb->curerrs >= max_errs) {
        Awrite(kdp->ident, msg_max_err, sizeof(msg_max_err) - 1);
        myfree(kdp);
    } else {
        EV_SET(&afds[used_afds++], kdp->ident, EVFILT_READ, EV_ADD | EV_ONESHOT, 0, 0, t_udb);
        EV_SET(&afds[used_afds++], kdp->ident, EVFILT_TIMER, EV_ADD, 0, max_timeout, t_udb);
    }
}

static void wel_cmd(const struct kevent *const kdp)
{
    int n;
    UDB *t_udb;

    t_udb = (UDB *) kdp->udata;
    n = Awrite(kdp->ident, msg_wel, sizeof(msg_wel) - 1);
    if (n == -1) {
        myfree(kdp);
        return;
    }
    t_udb->welmsg = 1;
    check_errs(kdp);
}

static void unknown_cmd(const struct kevent *const kdp)
{
    int n;
    UDB *t_udb;

    t_udb = (UDB *) kdp->udata;
    n = Awrite(kdp->ident, msg_unknown, sizeof(msg_unknown) - 1);
    if (n == -1) {
        myfree(kdp);
        return;
    }
    t_udb->curerrs++;
    t_udb->cmd = 0;
    t_udb->cmderr = 0;
    check_errs(kdp);
}

static void user_cmd(const struct kevent *const kdp)
{
    int n;
    UDB *t_udb;

    t_udb = (UDB *) kdp->udata;
    if (t_udb->cmderr != 0) {
        n = Awrite(kdp->ident, msg_user_err, sizeof(msg_user_err) - 1);
        t_udb->curerrs++;
    } else {
        n = Awrite(kdp->ident, msg_user_ok, sizeof(msg_user_ok) - 1);
    }
    if (n == -1) {
        myfree(kdp);
        return;
    }
    t_udb->cmd = 0;
    t_udb->cmderr = 0;
    check_errs(kdp);
}

static void pass_cmd(const struct kevent *const kdp)
{
    int n;
    UDB *t_udb;

    t_udb = (UDB *) kdp->udata;
    if (t_udb->cmderr == 0) {
        n = Awrite(kdp->ident, msg_pass_ok, sizeof(msg_pass_ok) - 1);
    } else if (t_udb->cmderr == 1) {
        n = Awrite(kdp->ident, msg_pass_no_user, sizeof(msg_pass_no_user) - 1);
        t_udb->curerrs++;
    } else if (t_udb->cmderr == 2) {
        n = Awrite(kdp->ident, msg_pass_no_pass, sizeof(msg_pass_no_pass) - 1);
        t_udb->curerrs++;
    } else {
        n = Awrite(kdp->ident, msg_pass_err, sizeof(msg_pass_err) - 1);
        t_udb->curerrs++;
    }
    if (n == -1) {
        myfree(kdp);
        return;
    }
    t_udb->cmd = 0;
    t_udb->cmderr = 0;
    check_errs(kdp);
}

static void noop_cmd(const struct kevent *const kdp)
{
    int n;
    UDB *t_udb;

    t_udb = (UDB *) kdp->udata;
    n = Awrite(kdp->ident, msg_noop, sizeof(msg_noop) - 1);
    if (n == -1) {
        myfree(kdp);
        return;
    }
    t_udb->cmd = 0;
    t_udb->cmderr = 0;
    check_errs(kdp);
}

static void quit_cmd(const struct kevent *const kdp)
{
    int n;

    n = Awrite(kdp->ident, msg_quit, sizeof(msg_quit) - 1);
    if (n == -1) {
        myfree(kdp);
        return;
    }
    myfree(kdp);
}

static void list_cmd(const struct kevent *const kdp)
{
    int i, n, usiz, tmp;
    char outbuf[512];
    UDB *t_udb;
    MINFO *t_cache;

    t_udb = (UDB *) kdp->udata;
    t_cache = t_udb->c_all;
    tmp = t_udb->a_id;
    if (tmp == -1) {
        n = Awrite(kdp->ident, msg_id_err, sizeof(msg_id_err) - 1);
        t_udb->cmderr++;
    } else if (tmp > t_udb->maxid) {
        n = Awrite(kdp->ident, msg_id_no, sizeof(msg_id_no) - 1);
        t_udb->cmderr++;
    } else if (t_udb->maxid == 0) {
        n = Awrite(kdp->ident, msg_list_no, sizeof(msg_list_no) - 1);
        t_udb->curerrs++;
    } else if (tmp == 0) {
        strlcpy(outbuf, msg_list_all, sizeof(outbuf));
        usiz = sizeof(msg_list_all) - 1;
        for (i=0; i<t_udb->maxid; i++) {
            if ((*(t_cache+i)).f_del != 1) {
                usiz = plcat(outbuf, usiz, sizeof(outbuf), "%u %u\r\n", i + 1, (*(t_cache+i)).size);
            }
            if (usiz < 0) {
                n = Awrite(kdp->ident, outbuf, -usiz);
                if (n == -1) {
                    myfree(kdp);
                    return;
                }
                usiz = 0;
                i--;
            } else if (i == t_udb->maxid - 1) {
                n = Awrite(kdp->ident, outbuf, usiz);
                if (n == -1) {
                    myfree(kdp);
                    return;
                }
            }
        }
        n = Awrite(kdp->ident, crlfend, sizeof(crlfend) - 1);
    } else if ((*(t_cache+tmp-1)).f_del == 1) {
        usiz = plcat(outbuf, 0, sizeof(outbuf), "-ERR message %u is marked as deleted\r\n", tmp);
        n = Awrite(kdp->ident, outbuf, usiz);
        t_udb->cmderr++;
    } else {
        usiz = plcat(outbuf, 0, sizeof(outbuf), "+OK %u %u\r\n", tmp, (*(t_cache+tmp-1)).size);
        n = Awrite(kdp->ident, outbuf, usiz);
    }
    if (n == -1) {
        myfree(kdp);
        return;
    }
    t_udb->cmd = 0;
    t_udb->cmderr = 0;
    check_errs(kdp);
}

static void uidl_cmd(const struct kevent *const kdp)
{
    int i, n, usiz, tmp;
    char outbuf[1024];
    UDB *t_udb;
    MINFO *t_cache;

    t_udb = (UDB *) kdp->udata;
    t_cache = t_udb->c_all;
    tmp = t_udb->a_id;
    if (tmp == -1) {
        n = Awrite(kdp->ident, msg_id_err, sizeof(msg_id_err) - 1);
        t_udb->cmderr++;
    } else if (tmp > t_udb->maxid) {
        n = Awrite(kdp->ident, msg_id_no, sizeof(msg_id_no) - 1);
        t_udb->cmderr++;
    } else if (t_udb->maxid == 0) {
        n = Awrite(kdp->ident, msg_uidl_no, sizeof(msg_uidl_no) - 1);
        t_udb->curerrs++;
    } else if (tmp == 0) {
        strlcpy(outbuf, msg_ok, sizeof(outbuf));
        usiz = sizeof(msg_ok) - 1;
        for (i=0; i<t_udb->maxid; i++) {
            if ((*(t_cache+i)).f_del != 1) {
                usiz = plcat(outbuf, usiz, sizeof(outbuf), "%u %s\r\n", i + 1, (*(t_cache+i)).name);
            }
            if (usiz < 0) {
                n = Awrite(kdp->ident, outbuf, -usiz);
                if (n == -1) {
                    myfree(kdp);
                    return;
                }
                usiz = 0;
                i--;
            } else if (i == t_udb->maxid - 1) {
                n = Awrite(kdp->ident, outbuf, usiz);
                if (n == -1) {
                    myfree(kdp);
                    return;
                }
            }
        }
        n = Awrite(kdp->ident, crlfend, sizeof(crlfend) - 1);
    } else if ((*(t_cache+tmp-1)).f_del == 1) {
        usiz = plcat(outbuf, 0, sizeof(outbuf), "-ERR message %u is marked as deleted\r\n", tmp);
        n = Awrite(kdp->ident, outbuf, usiz);
        t_udb->cmderr++;
    } else {
        usiz = plcat(outbuf, 0, sizeof(outbuf), "+OK %u %s\r\n", tmp, (*(t_cache+tmp-1)).name);
        n = Awrite(kdp->ident, outbuf, usiz);
    }
    if (n == -1) {
        myfree(kdp);
        return;
    }
    t_udb->cmd = 0;
    t_udb->cmderr = 0;
    check_errs(kdp);
}

static int cache_update(UDB *const t_udb, MINFO *const id_updated)
{
    int i;
    MINFO **p_updated;

    if (t_udb->c_updated == NULL) {
        t_udb->ucsize = 16;
        p_updated = (MINFO **) malloc(t_udb->ucsize * sizeof(MINFO *));
        if (p_updated == NULL) {
            INFO("cache_update: malloc(2) failed at cache_update: %s\n", strerror(errno));
            return (-1);
        }
        (MINFO **) t_udb->c_updated = (MINFO **) p_updated;
    }
    (MINFO **) p_updated = (MINFO **) t_udb->c_updated;
    i = 0;
    while (i < t_udb->upid) {
        if (*(p_updated+i) == id_updated) {
            return (0);
        } else {
            i++;
        }
    }
    if (i == t_udb->ucsize - 1) {
        t_udb->ucsize <<= 1;
        p_updated = (MINFO **) realloc(t_udb->c_updated, t_udb->ucsize * sizeof(MINFO *));
        if (p_updated == NULL) {
            INFO("cache_update: realloc(2) failed at cache_update: %s\n", strerror(errno));
            return (-1);
        }
        (MINFO **) t_udb->c_updated = (MINFO **) p_updated;
    }
    *(p_updated + t_udb->upid++) = id_updated;
    return (0);
}

static void dele_cmd(const struct kevent *const kdp)
{
    int n, tmp;
    char outbuf[64];
    UDB *t_udb;
    MINFO *t_cache;

    t_udb = (UDB *) kdp->udata;
    t_cache = t_udb->c_all;
    tmp = t_udb->a_id;
    if (tmp == -1) {
        n = Awrite(kdp->ident, msg_id_err, sizeof(msg_id_err) - 1);
        t_udb->cmderr++;
    } else if ((tmp > t_udb->maxid) || (t_udb->maxid == 0)) {
        n = Awrite(kdp->ident, msg_id_no, sizeof(msg_id_no) - 1);
        t_udb->cmderr++;
    } else if (tmp == 0) {
        n = Awrite(kdp->ident, msg_no_arg, sizeof(msg_no_arg) - 1);
        t_udb->cmderr++;
    } else if ((*(t_cache+tmp-1)).f_del == 1) {
        n = plcat(outbuf, 0, sizeof(outbuf), "-ERR message %u already marked as deleted\r\n", tmp);
        n = Awrite(kdp->ident, outbuf, n);
        t_udb->cmderr++;
    } else {
        (*(t_cache+tmp-1)).f_del = 1;
        n = cache_update(t_udb, t_cache+tmp-1);
        if (n == -1) {
            Awrite(kdp->ident, msg_dele_err, sizeof(msg_dele_err) - 1);
            myfree(kdp);
            return;
        }
        n = plcat(outbuf, 0, sizeof(outbuf), "+OK message %u marked as deleted\r\n", tmp);
        n = Awrite(kdp->ident, outbuf, n);
    }
    if (n == -1) {
        myfree(kdp);
        return;
    }
    t_udb->cmd = 0;
    t_udb->cmderr = 0;
    check_errs(kdp);
}

static void rset_cmd(const struct kevent *const kdp)
{
    int i, n;
    UDB *t_udb;
    MINFO **p_updated;

    t_udb = (UDB *) kdp->udata;
    if (t_udb->c_updated != NULL) {
        p_updated = (MINFO **) t_udb->c_updated;
        for (i=0; i<t_udb->upid; i++) {
            (*(p_updated+i))->f_read = 0;
            (*(p_updated+i))->f_del = 0;
        }
        free(t_udb->c_updated);
        t_udb->c_updated = NULL;
    }
    t_udb->ucsize = 0;
    t_udb->upid = 0;
    n = Awrite(kdp->ident, msg_rset, sizeof(msg_rset) - 1);
    if (n == -1) {
        myfree(kdp);
        return;
    }
    t_udb->cmd = 0;
    t_udb->cmderr = 0;
    check_errs(kdp);
}

static void stat_cmd(const struct kevent *const kdp)
{
    UDB *t_udb;
    MINFO *t_cache;
    int i, curmail, cursize;
    char outbuf[32];

    t_udb = (UDB *) kdp->udata;
    t_cache = t_udb->c_all;
    curmail = 0;
    cursize = 0;
    if (t_udb->maxid != 0) {
        for (i=0; i<t_udb->maxid; i++) {
            if ((*(t_cache+i)).f_del != 1) {
                cursize += (*(t_cache+i)).size;
                curmail++;
            }
        }
    }
    i = plcat(outbuf, 0, sizeof(outbuf), "+OK %u %u\r\n", curmail, cursize);
    i = Awrite(kdp->ident, outbuf, i);
    if (i == -1) {
        myfree(kdp);
        return;
    }
    t_udb->cmd = 0;
    t_udb->cmderr = 0;
    check_errs(kdp);
}

static void retr_cmd(const struct kevent *const kdp)
{
    UDB *t_udb;
    MINFO *t_cache;
    int n, tmp, ofd;
    char *psend, *mail;
    register int send_siz, mail_siz;
    char sendbuf[def_send_psiz], tmpbuf[64];

    t_udb = (UDB *) kdp->udata;
    t_cache = t_udb->c_all;
    tmp = t_udb->a_id;
    send_siz = (kdp->data > def_send_psiz) ? def_send_psiz : kdp->data;
    if (tmp == -1) {
        n = Awrite(kdp->ident, msg_id_err, sizeof(msg_id_err) - 1);
        t_udb->cmderr++;
        t_udb->cmd = 0;
    } else if (tmp > t_udb->maxid) {
        n = Awrite(kdp->ident, msg_id_no, sizeof(msg_id_no) - 1);
        t_udb->cmderr++;
        t_udb->cmd = 0;
    } else if (tmp == 0) {
        n = Awrite(kdp->ident, msg_no_arg, sizeof(msg_no_arg) - 1);
        t_udb->cmderr++;
        t_udb->cmd = 0;
    } else if ((*(t_cache+tmp-1)).f_del == 1) {
        n = plcat(tmpbuf, 0, sizeof(tmpbuf), "-ERR message %u already marked as deleted\r\n", tmp);
        n = Awrite(kdp->ident, tmpbuf, n);
        t_udb->cmderr++;
        t_udb->cmd = 0;
    } else {
        (*(t_cache+tmp-1)).f_read = 1;
        n = cache_update(t_udb, t_cache+tmp-1);
        if (n == -1) {
            myfree(kdp);
            return;
        }
        if ((t_udb->offset == 0L) && (t_udb->mail == NULL)){
            n = plcat(tmpbuf, 0, sizeof(tmpbuf), "+OK %u octets\r\n", (*(t_cache+tmp-1)).size);
            n = Awrite(kdp->ident, tmpbuf, n);
            if (n == -1) {
                myfree(kdp);
                return;
            }
        }
        if (t_udb->mail == NULL) {
            if ((*(t_cache+tmp-1)).f_new == 1) {
                (void)plcat(tmpbuf, 0, sizeof(tmpbuf), "%s/new/%s", t_udb->maildir, (*(t_cache+tmp-1)).name);
            } else {
                (void)plcat(tmpbuf, 0, sizeof(tmpbuf), "%s/cur/%s", t_udb->maildir, (*(t_cache+tmp-1)).name);
            }
            ofd = open(tmpbuf, O_RDONLY | O_NONBLOCK | O_DIRECT);
            if (ofd < 0) {
                INFO("retr_cmd: open(2) failed at retr_cmd: %s\n", strerror(errno));
                Awrite(kdp->ident, msg_open_err, sizeof(msg_open_err) - 1);
                myfree(kdp);
                return;
            }
            t_udb->mail = mmap(0, (*(t_cache+tmp-1)).size, PROT_READ, MAP_PRIVATE, ofd, 0L);
            close(ofd);
            if (t_udb->mail == MAP_FAILED) {
                INFO("retr_cmd: mmap(2) failed at retr_cmd: %s\n", strerror(errno));
                Awrite(kdp->ident, msg_mmap_err, sizeof(msg_mmap_err) - 1);
                myfree(kdp);
                return;
            }
        }
        psend = sendbuf;
        mail = t_udb->mail + t_udb->offset;
        mail_siz = (*(t_cache+tmp-1)).size;
        while ((psend - sendbuf + 1) <= send_siz - 1) {
            if (mail >= t_udb->mail + mail_siz) {
                break;
            }
            if (*mail == 0x0A) { // '\n'
                *psend++ = 0x0D; // '\r'
                *psend++ = 0x0A;
            } else if ((*mail  == 0x2E) && (*(mail-1) == 0x0A)) { // '.' , '\n'
                *psend++ = 0x2E; // '.'
                *psend++ = 0x2E; // '.'
            } else {
                *psend++ = *mail;
            }
            mail++;
        }
        if (((psend - sendbuf) < send_siz) && (mail < t_udb->mail + mail_siz)) {
            if (*mail == 0x0A) { // '\n'
                *psend = 0x00;
            } else if ((*mail  == 0x2E) && (*(mail-1) == 0x0A)) { // '.' , '\n'
                *psend = 0x00;
            } else {
                *psend++ = *mail;
            }
            mail++;
        }
        n = Write(kdp->ident, sendbuf, psend - sendbuf);
        if (n == -1) {
            myfree(kdp);
            return;
        }
        t_udb->offset = mail - t_udb->mail;
        if (t_udb->offset == mail_siz) {
            n = Awrite(kdp->ident, crlfend, sizeof(crlfend) - 1);
            n = munmap(t_udb->mail, mail_siz);
            if (n == -1) {
                INFO("retr_cmd: munmap(2) failed at retr_cmd: %s\n", strerror(errno));
                Awrite(kdp->ident, msg_munmap_err, sizeof(msg_munmap_err) - 1);
                myfree(kdp);
                return;
            }
            t_udb->cmd = 0;
            t_udb->mail = NULL;
            t_udb->offset = 0L;
        }
    }
    if (n == -1) {
        myfree(kdp);
        return;
    }
    if (t_udb->cmd == 0) {
        t_udb->cmderr = 0;
        check_errs(kdp);
    } else {
        EV_SET(&afds[used_afds++], kdp->ident, EVFILT_WRITE, EV_ADD | EV_ONESHOT, 0, 0, t_udb);
    }
}

static void exit_cmd(const struct kevent *const kdp)
{
    int i, res;
    char fdir[64], tdir[64];
    UDB *t_udb;
    MINFO **p_updated;

    t_udb = (UDB *) kdp->udata;
    if (t_udb->ucsize != 0) {
        (MINFO **) p_updated = (MINFO **) t_udb->c_updated;
        for (i=0; i<t_udb->upid; i++) {
            if ((*(p_updated+i))->f_del == 1) {
                if ((*(p_updated+i))->f_new == 1) {
                    (void) plcat(fdir, 0, sizeof(fdir), "%s/new/%s", t_udb->maildir, (*(p_updated+i))->name);
                } else {
                    (void) plcat(fdir, 0, sizeof(fdir), "%s/cur/%s", t_udb->maildir, (*(p_updated+i))->name);
                }
                res = unlink(fdir);
                if (res != 0) {
                    INFO("exit_cmd: unlink(2) failed at exit_cmd: %s\n", strerror(errno));
                    Awrite(kdp->ident, msg_unlink_err, sizeof(msg_unlink_err) - 1);
                    myfree(kdp);
                    return;
                }
            } else if ((*(p_updated+i))->f_read == 1) {
                if ((*(p_updated+i))->f_new == 1) {
                    (void) plcat(fdir, 0, sizeof(fdir), "%s/new/%s", t_udb->maildir, (*(p_updated+i))->name);
                    (void) plcat(tdir, 0, sizeof(tdir), "%s/cur/%s", t_udb->maildir, (*(p_updated+i))->name);
                    res = rename(fdir, tdir);
                    if (res != 0) {
                        INFO("exit_cmd: rename(2) failed at exit_cmd: %s\n", strerror(errno));
                        Awrite(kdp->ident, msg_rename_err, sizeof(msg_rename_err) - 1);
                        myfree(kdp);
                        return;
                    }
                }
            }
        }
    }
    i = Awrite(kdp->ident, msg_quit, sizeof(msg_quit) - 1);
    if (i == -1) {
        myfree(kdp);
        return;
    }
    myfree(kdp);
}

static void tmperr_cmd(const struct kevent *const kdp)
{
    int n;
    UDB *t_udb;

    t_udb = (UDB *) kdp->udata;
    n = Awrite(kdp->ident, msg_opendir_err, sizeof(msg_opendir_err) - 1);
    if (n == -1) {
        myfree(kdp);
        return;
    }
    t_udb->cmd = 0;
    t_udb->cmderr = 0;
    check_errs(kdp);
}

static void cmd_write(const struct kevent *const kdp)
{
    UDB *w_udb;

    w_udb = (UDB *) kdp->udata;
    if (w_udb->welmsg == 0) {
        wel_cmd(kdp);
    } else if (w_udb->cmd == DEF_CMDNAME_UNKNOWN) {
        unknown_cmd(kdp);
    } else if (w_udb->cmd == DEF_CMDNAME_USER) {
        user_cmd(kdp);
    } else if (w_udb->cmd == DEF_CMDNAME_PASS) {
        pass_cmd(kdp);
    } else if (w_udb->cmd == DEF_CMDNAME_QUIT) {
        quit_cmd(kdp);
    } else if (w_udb->cmd == DEF_CMDNAME_RSET) {
        rset_cmd(kdp);
    } else if (w_udb->cmd == DEF_CMDNAME_NOOP) {
        noop_cmd(kdp);
    } else if (w_udb->cmd == DEF_CMDNAME_EXIT) {
        exit_cmd(kdp);
    } else if (w_udb->c_status == DEF_CACHE_UNDONE) {
        tmperr_cmd(kdp);
    } else if (w_udb->cmd == DEF_CMDNAME_LIST) {
        list_cmd(kdp);
    } else if (w_udb->cmd == DEF_CMDNAME_UIDL) {
        uidl_cmd(kdp);
    } else if (w_udb->cmd == DEF_CMDNAME_DELE) {
        dele_cmd(kdp);
    } else if (w_udb->cmd == DEF_CMDNAME_RETR) {
        retr_cmd(kdp);
    } else if (w_udb->cmd == DEF_CMDNAME_STAT) {
        stat_cmd(kdp);
    }
}

static char *upch(char *const str)
{
    register char *tmp;

    tmp = str;
    while(*tmp != 0x00) {
        if ((*tmp > 0x7A) || (*tmp < 0x41)) { // 'z', 'A'
            return NULL;
        } else if ((*tmp > 0x5A) && (*tmp < 0x61)) { // 'Z', 'a'
            return NULL;
        } else if ((*tmp >= 0x61) && (*tmp <= 0x7A)) { // 'a', 'z'
            *tmp -= 32;
        }
        tmp++;
    }
    return (str);
}

static int aldigit(char *const data, const int maxn)
{
    int i;
    register char *t;

    if (*data == 0x00) { // '\0'
        return (0);
    }
    t = data;
    while (*t != 0x00) {
        if ((*t < 0x30) || (*t > 0x39)) { // '0', '9'
            return (-1);
        }
        t++;
    }
    i = atoi(data);
    return (i == 0) ? (maxn + 3) : (i);
}

static int cmd_parse(CMDTBL *const p_cmd, char *cmdbuf)
{
    int flags, values;
    char *tmp;

    tmp = cmdbuf + 4;
    if (*tmp != 0x0D && *tmp != 0x20) { // '\r', ' '
        return (0);
    }
    *tmp++ = 0x00; // '\0'
    cmdbuf = upch(cmdbuf);
    if (cmdbuf == NULL) {
        return (0);
    }
    strcpy(p_cmd->name, cmdbuf);
    if (*tmp == 0x0A) { // '\n'
        return (1);
    }
    flags = 0;
    values = 0;
    while(*tmp != 0x00) {
        if ((*tmp == 0x0D) && (*(tmp + 1) == 0x0A)) { // '\r', '\n'
            *tmp = 0x00; // '\0'
            flags = 1;
            break;
        } else if ((values == 0) && (*tmp != 0x20)) {
            values = 1;
        }
        tmp++;
    }
    if (flags == 0) {
        return (-1);
    }
    if (values == 1) {
        strcpy(p_cmd->value, cmdbuf + 5);
        return (2);
    } else {
        return (1);
    }
}

static int md_cache(const char *dir, int siz, const int new, UDB *const m_udb)
{
    int id;
    DIR *dirp;
    struct dirent *dp;
    MINFO *p;
    struct stat stat_buf;
    char tdir[64];

    if (m_udb->c_all == NULL) {
        p = (MINFO *) malloc(siz * sizeof(MINFO));
        if (p == NULL) {
            INFO("md_cache: malloc(2) failed at md_cache: %s\n", strerror(errno));
            return (-1);
        }
        (MINFO *) m_udb->c_all = p;
        (void) memset(p, 0, siz * sizeof(MINFO));
    }
    if ((dirp = opendir(dir)) == NULL) {
        INFO("md_cache: opendir(2) failed at md_cache: %s\n", strerror(errno));
        return (-1);
    }
    p = (MINFO *) m_udb->c_all;
    id = m_udb->maxid;
    while ((dp = readdir(dirp)) != NULL) {
        if (id == (siz - 1)) {
            siz <<= 1;
            p = (MINFO *) realloc(m_udb->c_all, siz * sizeof(MINFO));
            if (p == NULL) {
                INFO("md_cache: realloc(2) failed at md_cache: %s\n", strerror(errno));
                closedir(dirp);
                return (-1);
            }
            (MINFO *) m_udb->c_all = p;
        }
        if (((strcmp(".", dp->d_name)) == 0) || ((strcmp("..", dp->d_name)) == 0)) {
            continue;
        }
        (void)plcat(tdir, 0, sizeof(tdir), "%s/%s", dir, dp->d_name);
        if ((stat(tdir, &stat_buf)) == -1) {
            INFO("md_cache: stat(2) failed at md_cache: %s\n", strerror(errno));
            closedir(dirp);
            return (-1);
        }
        (*(p+id)).id = id + 1;
        (*(p+id)).f_del = 0;
        (*(p+id)).f_read = 0;
        (*(p+id)).f_new = new;
        strlcpy((*(p+id)).name, dp->d_name, sizeof(p->name));
        (*(p+id)).size = stat_buf.st_size;
        id++;
    }
    m_udb->maxid = id;
    if ((closedir(dirp)) != 0) {
        INFO("md_cache: closedir(3) failed at md_cache: %s\n", strerror(errno));
        return (-1);
    }
    return (siz);
}

static void init_cache(UDB *const g_udb)
{
    register int res;
    char tmpdir[64];

    (void) plcat(tmpdir, 0, sizeof(tmpdir), "%s/new", g_udb->maildir);
    res = md_cache(tmpdir, 32, 1, g_udb);
    if (res == -1) {
        return;
    }
    (void) plcat(tmpdir, 0, sizeof(tmpdir), "%s/cur", g_udb->maildir);
    res = md_cache(tmpdir, res, 0, g_udb);
    if (res == -1) {
        return;
    }
    g_udb->c_status = DEF_CACHE_DONE;
}

static void cmd_read(const struct kevent *const kdp)
{
    int n;
    char inbuf[48];
    UDB *r_udb;
    CMDTBL r_cmd;

    r_udb = (UDB *) kdp->udata;
    if (r_udb->free <= DEF_UDB_FREEED) {
        r_udb->free = DEF_UDB_INIT;
        return;
    }
    n = read(kdp->ident, &inbuf, sizeof(inbuf)-1);
    if (n == -1) {
        INFO("cmd_read: read(2) socket error: %s\n", strerror(errno));
        myfree(kdp);
        return;
    } else if (n == 0) {
        INFO("cmd_read: read(2) socket error: %s\n", strerror(errno));
        myfree(kdp);
        return;
    } else {
        *(inbuf+n) = 0x00; // '\0'
        (void) memset(&r_cmd, 0, sizeof(CMDTBL));
        n = cmd_parse(&r_cmd, inbuf);
        if (n == -1) {
            Awrite(kdp->ident, msg_over, sizeof(msg_over) - 1);
            myfree(kdp);
            return;
        } else if (n == 0) {
            r_udb->cmd = DEF_CMDNAME_UNKNOWN;
            r_udb->cmderr = 0;
        } else if (r_udb->login == 0) {
            if ((strncmp(r_cmd.name, "USER", 4)) == 0) {
                r_udb->cmd = DEF_CMDNAME_USER;
                if (n == 2) {
                    (void) plcat(r_udb->username, 0, sizeof(r_udb->username), "%s", r_cmd.value);
                    r_udb->cmderr = 0;
                } else {
                    r_udb->cmderr = 1;
                }
            } else if ((strncmp(r_cmd.name, "PASS", 4)) == 0) {
                r_udb->cmd = DEF_CMDNAME_PASS;
                if (r_udb->username[0] == '\0') {
                    r_udb->cmderr = 1;
                } else if (n == 1) {
                    r_udb->cmderr = 2;
                } else {
                    n = auth_system(r_udb->username, r_udb->maildir, sizeof(r_udb->maildir), r_cmd.value);
                    if (n == 1) {
                        r_udb->cmderr = 0;
                        r_udb->login = 1;
                    } else {
                        r_udb->cmderr = 3;
                    }
                }
            } else if ((strncmp(r_cmd.name, "QUIT", 4)) == 0) {
                r_udb->cmd = DEF_CMDNAME_QUIT;
                r_udb->cmderr = 0;
            } else {
                r_udb->cmd = 0;
            }
        } else {
            if ((strncmp(r_cmd.name, "LIST", 4)) == 0) {
                r_udb->cmd = DEF_CMDNAME_LIST;
                r_udb->a_id = aldigit(r_cmd.value, r_udb->maxid);
            } else if ((strncmp(r_cmd.name, "UIDL", 4)) == 0) {
                r_udb->cmd = DEF_CMDNAME_UIDL;
                r_udb->a_id = aldigit(r_cmd.value, r_udb->maxid);
            } else if ((strncmp(r_cmd.name, "DELE", 4)) == 0) {
                r_udb->cmd = DEF_CMDNAME_DELE;
                r_udb->a_id = aldigit(r_cmd.value, r_udb->maxid);
            } else if ((strncmp(r_cmd.name, "RETR", 4)) == 0) {
                r_udb->cmd = DEF_CMDNAME_RETR;
                r_udb->a_id = aldigit(r_cmd.value, r_udb->maxid);
            } else if ((strncmp(r_cmd.name, "STAT", 4)) == 0) {
                r_udb->cmd = DEF_CMDNAME_STAT;
                r_udb->cmderr = 0;
            } else if ((strncmp(r_cmd.name, "RSET", 4)) == 0) {
                r_udb->cmd = DEF_CMDNAME_RSET;
                r_udb->cmderr = 0;
            } else if ((strncmp(r_cmd.name, "NOOP", 4)) == 0) {
                r_udb->cmd = DEF_CMDNAME_NOOP;
                r_udb->cmderr = 0;
            } else if ((strncmp(r_cmd.name, "QUIT", 4)) == 0) {
                r_udb->cmd = DEF_CMDNAME_EXIT;
                r_udb->cmderr = 0;
            } else {
                r_udb->cmd = DEF_CMDNAME_UNKNOWN;
                r_udb->cmderr = 0;
            }
            if ((r_udb->cmd < DEF_CMDNAME_RSET) && (r_udb->c_status == DEF_CACHE_UNDONE)) {
                init_cache(r_udb);
            }
        }
        EV_SET(&afds[used_afds++], kdp->ident, EVFILT_WRITE, EV_ADD | EV_ONESHOT, 0, 0, r_udb);
        EV_SET(&afds[used_afds++], kdp->ident, EVFILT_TIMER, EV_DELETE, 0, 0, NULL);
    }
}

static void timeout_cmd(const struct kevent *const kdp)
{
    int n;
    UDB *t_udb;

    t_udb = (UDB *) kdp->udata;
    n = Awrite(kdp->ident, msg_timeout, sizeof(msg_timeout) - 1);
    myfree(kdp);
}

static void checktm(const struct kevent *const kdp)
{
    UDB *t_udb;

    t_udb = (UDB *) kdp->udata;
    if (t_udb->free >= DEF_UDB_USED) {
        timeout_cmd(kdp);
    } else {
        t_udb->free = DEF_UDB_INIT;
    }
}

static void myloop(void)
{
    int res;
    struct kevent *kdp;

    while (1) {
        res = kevent(kq, afds, used_afds, aevs, max_client, NULL);
        used_afds = 0;
        if (res == -1) {
            INFO("myloop: kevent(2) failed at myloop: %s\n", strerror(errno));
            continue;
        }
        if (res == 0) {
            INFO("myloop: NO Event received\n");
            exitfree();
            _exit(EXIT_FAILURE);
        }
        for (kdp = aevs; kdp < &aevs[res]; kdp++) {
            if (kdp->flags & EV_ERROR) {
                continue;
            } else if (kdp->ident == pop3d_fd) {
                cli_accept(kdp);
            } else if (kdp->flags & EV_EOF) {
                myfree(kdp);
            } else if (kdp->filter == EVFILT_TIMER) {
                checktm(kdp);
            } else if (kdp->filter == EVFILT_READ) {
                cmd_read(kdp);
            } else if (kdp->filter == EVFILT_WRITE) {
                cmd_write(kdp);
            }
        }
        used_fid = 0;
    }
}

void pop3d(void)
{
    if (kq_init() == -1) {
        exitfree();
        _exit(EXIT_FAILURE);
    }
    EV_SET(&afds[used_afds++], pop3d_fd, EVFILT_READ, EV_ADD, 0, 0, NULL);
    myloop();
}
